/* ipaudit.c
 *
 * ipaudit - network traffic data gathering
 * By Jon Rifkin <jon.rifkin@uconn.edu>
 * Copyright 1999,2000 Jonathan Rifkin
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


/*
------------------------------------------------------------------------
Compile Switches
------------------------------------------------------------------------
*/
#define DEBUG

#define DUMP
#undef  DUMP

#define HASH
#undef  HASH


/*
------------------------------------------------------------------------
Include Files
------------------------------------------------------------------------
*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <pcap.h>
#include <signal.h>
#include <limits.h>
#include <unistd.h>
#include <time.h>
#include "hash.h"



/*
------------------------------------------------------------------------
Defines
------------------------------------------------------------------------
*/
#define VERSION_STR "ipaudit 0.93b3"

#define TRUE 1
#define FALSE 0

#if 0
/*  TCP ports to dump  */
#define NPORT           6

#define P_FTP          21
#define P_TELNET       23
#define P_SUNRPC      111
#define P_LOGIN       513
#define P_INGRESLOCK 1524
#define P_IRC        6667

/*  Misc protocols to dump  */
#define PROT_ICMP       1
#endif

/*  Protocols with port info  */
#define PROT_TCP        6
#define PROT_UDP       17


/*  Flags for udp/tcp accepting all/some ports  */
#define PROT_ACC_ALL    1
#define PROT_ACC_SOME   2

/*  Length of saved packets  */
#define PLEN_DEF 96  /* default  */
#define PLEN_MIN 68  /* min allowed  */

/*  
Length of packet headers  
(culled this info from tcpdump source code)
*/
#define POFF_ETH  14
#define POFF_NULL  4   /* Used by loopback ?  */
#define POFF_PPP   4
#define POFF_RAW   0

/*  
Number of hash slots
NOT number of packets, they're unlimited (except for memory)
*/
#define N_HASH_SLOTS 1000000

/*  Number of 1/10,000 of second in second  */
#define M0SEC 10000

/*
------------------------------------------------------------------------
DEBUGGING MACROS
------------------------------------------------------------------------
*/
#ifdef DEBUG
#define WRITEMSG \
	if (debug_m) { \
	printf ("File %s line %d: ", __FILE__, __LINE__); \
	printf ("errmsg <%s>\n", strerror(errno)); fflush(stdout); \
	}
#define WRITEVAR(VAL,FMT) \
	if (debug_m) { \
	printf ("File %s line %d: ", __FILE__, __LINE__); \
	printf ("%s=",#VAL); printf (#FMT, VAL); printf ("\n"); \
	fflush(stdout); \
	}
#else
#define WRITEMSG
#define WRITEVAR(VAL,FMT)
#endif



/*
------------------------------------------------------------------------
MACROS
------------------------------------------------------------------------
*/
/*  Convert time in 1/M0SEC to hours, min, seconds, 1/M0SEC  */
#define HMS(hour,min,sec,sec4) \
	sec   = sec4/M0SEC; \
	sec4 -= M0SEC*sec; \
	min   = sec/60; \
	sec  -= 60*min; \
	hour  = min/60; \
	min  -= 60*hour;


/*
------------------------------------------------------------------------
Type Definitions
------------------------------------------------------------------------
*/

/*  Packet structure used by pcap library  */
typedef struct {
	u_char src[6];
	u_char dst[6];
	u_char ptype[2];     /*  ==0x800 if ip  */
	u_char version[1];
	u_char service[1];
	u_char length[2];
	u_char id[2];
	u_char flag[2];
	u_char ttl[1];
	u_char prot[1];
	u_char chksum[2];
	u_char srcip[4];
	u_char dstip[4];
	u_char srcpt[2];
	u_char dstpt[2];
	} pkt_struct_t;

typedef struct {
	u_char src[6];
	u_char dst[6];
	u_char ptype[2];     /*  ==0x800 if ip  */
	} eth_struct_t;

typedef struct {
	u_char version[1];
	u_char service[1];
	u_char length[2];
	u_char id[2];
	u_char flag[2];
	u_char ttl[1];
	u_char prot[1];
	u_char chksum[2];
	u_char srcip[4];
	u_char dstip[4];
	u_char srcpt[2];
	u_char dstpt[2];
	} ip_struct_t;



/*  Start and stop time of each connection  */
typedef struct {
	/*  Time (in sec/10,000) of first and last packet  */
	int first_time, last_time;
	/*  Indentity of machine source for first, last packet  */
	unsigned char first_mach, last_mach;
} datatime_t;


/*  All data for connection  */
typedef struct {
	long       nbyte1, nbyte2;
	int        npkt1, npkt2;
	datatime_t time;
} data_t;







/*
------------------------------------------------------------------------
Global variables
------------------------------------------------------------------------
*/
extern int errno;
extern char pcap_version[];



/*
------------------------------------------------------------------------
Module variables
------------------------------------------------------------------------
*/
int isig_m=0;

u_char *prots_m = NULL;

u_char *tcp_ports_m = NULL;
u_char *udp_ports_m = NULL;


/*  Flag for writing connection time in output  */
int write_time_m = FALSE;
int printeth_m   = FALSE;

int  debug_m     = FALSE;

/*  Pcap input file  */
pcap_t *pcapfile_m = NULL;

int npkt_m = 0;      /*  Number of    packets  */
int nippkt_m  = 0;   /*  Number of ip packets  */
int nconn_m   = 0;   /*  Number of connections */

u_char ip_m[4] = "";

/*
------------------------------------------------------------------------
Local Function Prototypes
------------------------------------------------------------------------
*/
void ihandler (int);
void dum2 (pkt_struct_t *);
void dum1 (struct pcap_pkthdr *);
void dum3 (hlist_t **);
void storepkt 
   (struct pcap_pkthdr *, eth_struct_t *, ip_struct_t *, hlist_t **);
void writepkttxt (hlist_t **, char *, char *);
void writepktbin (hlist_t **, char *, char *);
void PrintUsage();

void parse_portstr(char *str);
void add_protocol(int val);
void add_port (int prot, int port);

void *alloc_err (int, int);
char *get_localstr (char *);

int  get_pkttime (struct pcap_pkthdr *);
int  get_packetoffset (int);


int cmptime (const void *, const void *);
int cmpip   (const void *, const void *);


/*
------------------------------------------------------------------------
Main Function
------------------------------------------------------------------------
*/
int main (int argc, char *argv[]) {
	char   ebuf[PCAP_ERRBUF_SIZE];
	struct pcap_pkthdr pkthdr;
	u_char       *raw_pkt = NULL;
	eth_struct_t *eth_pkt = NULL;
	ip_struct_t  *ip_pkt  = NULL;
	struct bpf_program  fcode;
	pcap_dumper_t *df = NULL;
	int dump_this = FALSE;
	int ival;
	int length;
	int i;
	u_char *dumptable = NULL;
	int  optchar;
	int  outputbin  = FALSE;
	int  promisc    = 1;          /*  Default, set promiscuius mode */
	FILE *pidfile   = NULL;
	char *progarg   = NULL;
	char *progfile  = NULL;
	char *writefile = NULL;
	char *readfile  = NULL;
	char *outfile   = NULL;
	int  maxpkt     = 0;
	int  nlen       = PLEN_DEF;   /*  Packet length to dump  */
	char *localstr  = NULL;	      /*  Sorting order for ip address */
	/*  Hash table for ip connections  */
	hlist_t **hconn = NULL;
	int  hostonly  = FALSE;
	int  uselimit  = FALSE;
	int  hostportlimit = 0;
	int  hostlimit     = 0;
	u_char nullip[4] = {0,0,0,0};
	int  DataLinkType;
	int  PacketOffset;
	char *filtercmd = "";
	int  nhashslots = N_HASH_SLOTS;
	int  IsEthernet = FALSE;
	

	/*  Read options  */
	while (-1 != (optchar=getopt(argc,argv,"HI:L:N:abc:dei:f:l:mo:p:r:s:tx:w:v"))) {
		switch (optchar) {
			case '?':
			return 1;
		case 'v':
			printf ("%s (compiled %s)\n", VERSION_STR, __DATE__);
			printf ("libpcap version %s\n", pcap_version);
			printf ("Default number of hash slots = %d\n", N_HASH_SLOTS);
			return 0;
		/*  Output packet statistics in binary file format  */
		case 'b':
			outputbin = TRUE;
			break;
		/*  Debugging option  */
		case 'd':
			printf ("%s (compiled %s)\n", VERSION_STR, __DATE__);
			printf ("libpcap version %s\n", pcap_version);
			printf ("Default number of hash slots = %d\n", N_HASH_SLOTS);
			debug_m = TRUE;
			break;
		/*  Print ethernet addresses if present  */
		case 'e':
			printeth_m = TRUE;
			break;
		/*  Get pcap filter string  */
		case 'f':
			filtercmd = strdup (optarg);
			break;
		/*  Write pid file  */
		case 'i':
			pidfile = fopen(optarg,"wt");
			if (NULL==pidfile) {
				fprintf (stderr, "ERROR:  Cannot open pidfile '%s'\n", optarg);
				exit(1);
			}
			fprintf (pidfile, "%d\n", getpid());
			fclose (pidfile);
			break;
		case 'x':
			progfile = strdup(optarg);
			break;
		case 'r':
			readfile = strdup(optarg);
			break;
		case 'w':
			writefile = strdup(optarg);
			break;
		case 'c':
			maxpkt    = atoi(optarg);
			break;
		case 'o':
			outfile   = strdup(optarg);
			break;
		case 'm':
			promisc   = 0;
			break;
		case 'p':
			parse_portstr(strdup(optarg));
			break;
		case 's':
			nlen = atoi(optarg);
			if (nlen<PLEN_MIN)  
				nlen = PLEN_MIN;
			break;
		case 'l':
			localstr = get_localstr (optarg);
			break;
		case 't':
			write_time_m = TRUE;
			break;
		case 'H':
			hostonly = TRUE;
			break;
		case 'I':
			sscanf (optarg, "%u.%u.%u.%u", &ip_m[0],&ip_m[1],&ip_m[2],&ip_m[3]);
			break;
		case 'L':
			sscanf (optarg, "%d,%d", &hostportlimit, &hostlimit);
			uselimit = TRUE;
			break;
		case 'N':
			nhashslots = atoi(optarg);
			if (nhashslots<1) {
				printf (
				"ERROR: Number of hash slots (-N%d) must be one or greater\n", 
				nhashslots);
				exit(1);
			}
			break;
		default:
			return 1;
		}
	}
	
	
	
	/*  If not reading pcap file need the interface name  */
	if (!readfile && argc-optind<1)  {
		PrintUsage();
		return 0;
	}
	

	/*  Open pcap file  */
	if (readfile) {
		pcapfile_m = pcap_open_offline(readfile, ebuf);
		if (NULL==pcapfile_m) {
			fprintf (stderr, "ERROR:  Cannot open read file %s.\n", readfile);
			exit(1);
		}
	} else {
		pcapfile_m = pcap_open_live(argv[optind], nlen, promisc, 1000, ebuf);
		if (pcapfile_m==NULL) {
			printf("ipaudit: %s (%s)\n", ebuf, "Do you need root?");
			return 1;
		}
	}

	/*  
	Read datalink type  
	This tells us if the datalink is ethernet, ppp, etc.
	*/
	DataLinkType = pcap_datalink(pcapfile_m);

	/*  Find packet header offset */
	PacketOffset = get_packetoffset (DataLinkType);

	IsEthernet = (DataLinkType==DLT_EN10MB || DataLinkType==DLT_IEEE802);


if (debug_m) {
	switch (DataLinkType) {
		case DLT_EN10MB: 
			printf ("DataLinkType = %s\n", "DLT_EN10MB"); break;
		case DLT_IEEE802: 
			printf ("DataLinkType = %s\n", "DLT_IEEE802"); break;
		case DLT_SLIP: 
			printf ("DataLinkType = %s\n", "DLT_SLIP"); break;
		case DLT_SLIP_BSDOS: 
			printf ("DataLinkType = %s\n", "DLT_SLIP_BSDOS"); break;
		case DLT_PPP: 
			printf ("DataLinkType = %s\n", "DLT_PPP"); break;
		case DLT_PPP_BSDOS: 
			printf ("DataLinkType = %s\n", "DLT_PPP_BSDOS"); break;
		case DLT_FDDI: 
			printf ("DataLinkType = %s\n", "DLT_FDDI"); break;
		case DLT_NULL: 
			printf ("DataLinkType = %s\n", "DLT_NULL"); break;
		case DLT_RAW: 
			printf ("DataLinkType = %s\n", "DLT_RAW"); break;
		case DLT_ATM_RFC1483: 
			printf ("DataLinkType = %s\n", "DLT_ATM_RFC1483"); break;
		default:
			printf ("DataLinkType = %d\n", DataLinkType);
	}
}

   if (debug_m) 
	   printf ("PacketOffset = %d\n", PacketOffset);

	/*  
	Insure if ethernet addresses requested that
	captured packets contain them
	*/
	if (printeth_m && PacketOffset<0) {
		fprintf (stderr, 
			"ERROR:  Cannot print ethernet addresses as requested.\n");
		fprintf (stderr, "Current network interface (%s) ", argv[optind]);
		fprintf (stderr, "is not an ethernet interface.\n");
		exit(2);
	}

	/*  
	Apply user requested packet filter code
	*/
	if (pcap_compile(pcapfile_m, &fcode, filtercmd, 0, 0) < 0)
		printf("compile: %s", pcap_geterr(pcapfile_m));
	if (pcap_setfilter(pcapfile_m, &fcode) < 0)
		printf("setfilter:  %s", pcap_geterr(pcapfile_m));

	/*  Problem with pcap_setfilter?  Sets error, unset here  */
	errno = 0;


	/*  Install interupt handler if reading live */
	if (!readfile) {
		signal (SIGINT,    ihandler);   /*  intercepts  ^C           */
		signal (SIGTERM,   ihandler);   /*  intercepts  ^kill <PID>  */
	}

	/*  Initialize hash table  */
	hconn = init_htable(N_HASH_SLOTS);


	/*  Counters  */
	npkt_m    = 0;
	nippkt_m  = 0;

	/*  Open dump file(s)  */
	if (writefile  ) df     = pcap_dump_open(pcapfile_m, writefile);


	/*  Read inteface until interupt signal  */
	while (isig_m==0) {
		npkt_m++;
		raw_pkt = (u_char *) pcap_next (pcapfile_m, &pkthdr);

		/*  If pkt is null and we're reading file then we're done  */
		/*  Otherwise if reading live try again                    */
		if (raw_pkt==NULL) {
			if (readfile) break;
			else          continue;
		}

		/*  Skip this packet if ethernet and not ip  */
		if (IsEthernet) {
			eth_pkt = (eth_struct_t *) raw_pkt;
			if (! (eth_pkt->ptype[0]==8 && eth_pkt->ptype[1]==0) ) continue;
		}

		/*  Find pointer to ip packet  */
		ip_pkt = (ip_struct_t *)( raw_pkt + PacketOffset );


		/*  Don't exceed limit of ip packets  */
		nippkt_m++;
		if (maxpkt && nippkt_m>maxpkt)
			break;

/*  Dump packet contents  */
if (debug_m) {
	int ibyte, iwidth;
	printf ("IP Packet Count   %d\n", nippkt_m);
	printf ("Raw Packet Length %d\n", pkthdr.len);
	printf ("Captured   Length %d\n", pkthdr.caplen);
	printf ("Captured bytes ...\n");
	iwidth=0;
	for (ibyte=0;ibyte<pkthdr.caplen;ibyte++) {
		printf (" %03d", raw_pkt[ibyte]);
		if (++iwidth==16) {
			printf ("\n");
			iwidth=0;
		}
	}
	printf ("\n");
}

		/*  Host only storage, set prot, port to zero  */
		if (hostonly) {
			memset (ip_pkt->srcpt, 0, 2);
			memset (ip_pkt->dstpt, 0, 2);
			memset (ip_pkt->prot,  0, 1);
		}

		/*  If number of stored packets with full host,port info exceeded
		then just store host info.  If number of packets with host info
		exceeded, then just increment byte and packet info with a dummy
		host pair 0.0.0.0 0.0.0.0   */


		/*  Exceeded host/port limit  */
		if (uselimit && nconn_m >= hostportlimit) {

			/*  Set both host ports to 0  */
			memset (ip_pkt->srcpt, 0, 2);
			memset (ip_pkt->dstpt, 0, 2);

			/*  Exceeed host-only limit also  */
			if (nconn_m >= hostportlimit + hostlimit) {

				/*  Set both host IPs to 0.0.0.0  */
				memset (ip_pkt->srcip, 0, 4);
				memset (ip_pkt->dstip, 0, 4);
				if (eth_pkt) {
					memset (eth_pkt->src, 0, 6);
					memset (eth_pkt->dst, 0, 6);
				}
			}
		}
				

		/*  Set ports to 0 if not UDP or TCP  */
		if ( ip_pkt->prot[0]!=0x11 && ip_pkt->prot[0]!=0x06 ) {
			memset (ip_pkt->srcpt, 0, 2);
			memset (ip_pkt->dstpt, 0, 2);
		}

if (debug_m) {
printf ("%03d.%03d.%03d.%03d %03d.%03d.%03d.%03d  %3d %5d %5d\n", 
	ip_pkt->srcip[0],ip_pkt->srcip[1],ip_pkt->srcip[2],ip_pkt->srcip[3],
	ip_pkt->dstip[0],ip_pkt->dstip[1],ip_pkt->dstip[2],ip_pkt->dstip[3],
	ip_pkt->prot[0],
	ip_pkt->srcpt[0]*256+ip_pkt->srcpt[1],
	ip_pkt->dstpt[0]*256+ip_pkt->dstpt[1]);
}

		/*  Store packets  */
		storepkt (&pkthdr, eth_pkt, ip_pkt, hconn);


		/*  Using dump  */
		if (writefile) {

			/*  No protocols/ports specified, so dump all */
			dump_this = FALSE;
			if (NULL==prots_m && NULL==ip_m) 
				dump_this = TRUE;

			
			/*  Is this packet correct protocol/port  */
			if (prots_m) {
				dump_this = prots_m[ip_pkt->prot[0]];

				/*  If udp or tcp, are ports specified ?  */
				if (PROT_ACC_SOME==dump_this) {
					dumptable = (PROT_TCP==ip_pkt->prot[0]) ? tcp_ports_m : udp_ports_m;
					dump_this = 
						dumptable[(ip_pkt->srcpt[0]<<8)+ip_pkt->srcpt[1]] ||
						dumptable[(ip_pkt->dstpt[0]<<8)+ip_pkt->dstpt[1]];
				}
			} 

			if (!dump_this && ip_m) {
				dump_this = (!memcmp(ip_m,ip_pkt->srcip,4) || !memcmp(ip_m,ip_pkt->dstip,4));
			}

			/*  Dump packet  */
			if (dump_this)
				pcap_dump ((u_char *) df, &pkthdr, raw_pkt);
		}

		


	}  /*  Read tcpdump data  */

	if (NULL!=pcapfile_m)  pcap_close(pcapfile_m);

	/*  Clear error if breaking during pcap call  */
	errno = 0;


	/*  Close dump file  */
	if (writefile  ) pcap_dump_close (df);

	/*  Write packet info  */
	if (outputbin) {
		writepktbin(hconn, outfile, localstr);
	} else {
		writepkttxt(hconn, outfile, localstr);
	}

	/*  Call next program  */
	if (progfile) {
		progarg = index (progfile, ' ');
		if (progarg) {
		*progarg++ = '\0';
		}
		execl (progfile, progfile, progarg, NULL);
	}
}



/*
Interupt handler (called when program recieves operating system signal
*/
void ihandler (int cursig) {

	/*  Close pcap input  */
	pcap_close (pcapfile_m);
	pcapfile_m = NULL;

	/*  SET SIGNAL FLAG  */
	isig_m = 1;

	/*  FLUSH BUFFERS  */
	fflush (stdout);

	/*  RE-INSTALL SIGNAL HANDLER  */
	signal (cursig, SIG_DFL);
#if 0
	signal (cursig, ihandler);
#endif
#ifdef DEBUG
	if (debug_m) {
		struct tm *tm;
		time_t    seconds;
		fprintf (stderr, "ipaudit received signal number <%i>\n", cursig);
		time (&seconds);
		tm = localtime (&seconds);
		fprintf (stderr, "date is <%04d-%02d-%02d-%02d:%02d:%02d>\n", 
		  tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday, 
		  tm->tm_hour, tm->tm_min, tm->tm_sec);
		fprintf (stderr, "number of     packets read <%d>\n", npkt_m);
		fprintf (stderr, "number of ip  packets read <%d>\n", nippkt_m);
		fprintf (stderr, "number of connections      <%d>\n", getcount());
		}
#endif
	}


/*
Store packet info in hash table, 
keyed by ip1,ip2,port1,por2,protocol
data  is number of incoming/outgoing bytes, packets
*/
void storepkt (
struct pcap_pkthdr *pkthdr, 
eth_struct_t *ep, 
ip_struct_t *ip, 
hlist_t **ha
) {
	u_char     key[25];  /*  space for ip1,ip2,prot,prt1,prt2,eth1,eth2  */
	data_t     *data;
	int        ndata;
	int        length;
	data_t     idata;
	datatime_t idatatime;
	int        datasize;
	int        keysize;

	/*  Calculate data packet length  */
	length = ip->length[1] + 256*(int) ip->length[0] + 14;

	/*  
	Make key - order so smallest ip first 
	store data
	*/
	if (memcmp(ip->srcip, ip->dstip, 4) < 0) {
		memcpy (key+ 0, ip->srcip, 4);
		memcpy (key+ 4, ip->dstip, 4);
		memcpy (key+ 8, ip->srcpt, 2);
		memcpy (key+10, ip->dstpt, 2);
		if (ep) {
			memcpy (key+13, ep->src,   6);
			memcpy (key+19, ep->dst,   6);
		}
		idata.nbyte1 = 0;
		idata.nbyte2 = length;
		idata.npkt1  = 0;
		idata.npkt2  = 1;
	} else {
		memcpy (key+ 0, ip->dstip, 4);
		memcpy (key+ 4, ip->srcip, 4);
		memcpy (key+ 8, ip->dstpt, 2);
		memcpy (key+10, ip->srcpt, 2);
		if (ep) {
			memcpy (key+13, ep->dst,   6);
			memcpy (key+19, ep->src,   6);
		}
		idata.nbyte1 = length;
		idata.nbyte2 = 0;
		idata.npkt1  = 1;
		idata.npkt2  = 0;
	}
	memcpy (key+12, ip->prot,  1);


	/*  Store time if requested  */
	if (write_time_m) {
		idata.time.first_time = get_pkttime (pkthdr);
		/*  If machine 1 received packet, then source was 2, else 1  */
		idata.time.first_mach = (idata.npkt1==1) ? 2 : 1;
		/*  Set first and last machine info the same  */
		idata.time.last_time  = idata.time.first_time;
		idata.time.last_mach  = idata.time.first_mach;
		/*  Set size of full data structure  */
		datasize = sizeof(data_t);

	/*  Set datasize to not store time (saves memory space)  */
	} else {
		datasize = sizeof(data_t) - sizeof(datatime_t);
	}

#ifdef DUMP
printf ("%03u.%03u.%03u.%03u ", key[0], key[1], key[2], key[3]);
printf ("%03u.%03u.%03u.%03u ", key[4], key[5], key[6], key[7]);
printf ("%u %u %u %d %d\n", key[12], (int )key[8]*256+key[9], (int )key[10]*256+key[11], length, 1);
#endif

	/*  Set keysize according to whether we are storing eth packets  */
	if (printeth_m) {
		keysize = sizeof(key);
	} else {
		keysize = sizeof(key)-12;
	}

	/*  Add to table  */
	if (! finddata(ha,(u_char *)&key, keysize, (u_char **)&data,&ndata) ) {
		datasize = sizeof(idata);
		addnode (ha, (u_char *) &key, keysize, (u_char *) &idata, datasize);
		nconn_m++;  /*  Increment number of connections  */
	} else {
		data->nbyte1 += idata.nbyte1;
		data->nbyte2 += idata.nbyte2;
		data->npkt1  += idata.npkt1;
		data->npkt2  += idata.npkt2;
		/*  Update last packet time  */
		if (write_time_m) {
			data->time.last_time  = idata.time.last_time  ;
			data->time.last_mach  = idata.time.last_mach;
		}
#ifdef DUMP
printf ("data idata  <%u %u %u %u>a  <%u %u %u %u>\n", 
data[0], data[1], data[2], data[3],
idata[0], idata[1], idata[2], idata[3]);
#endif
	}
}


/*
Retrieve and print packets from hash table in bin format
*/
void writepktbin (hlist_t **ha, char *outname, char *localstr) {
	hlist_t *t;
	data_t  *data;
	FILE    *outfile = stdout;
	char    ip1[16], ip2[16];
	int     switch_mach;
	int     first_mach, last_mach;

	/*  Open file if outname is no
	if ( NULL!=outname && strcmp("-",outname) )
		outfile = fopen (outname, "wb");

	if (NULL==outfile) {
		fprintf (stderr, "ERROR:  Cannot open output file <%s>\n", outname);
		exit(1);
	}

	/*  Walk list  */
	getfirst (ha);
	while ((t=getnext(ha))) {

		/*  Get ip addresses and ports  */
		sprintf (ip1, "%03u.%03u.%03u.%03u", 
			t->key[0], t->key[1], t->key[2], t->key[3]);
		sprintf (ip2, "%03u.%03u.%03u.%03u", 
			t->key[4], t->key[5], t->key[6], t->key[7]);
		data = (data_t *) t->data;

		/*  Re-order ip addresses if 2nd is local and first is not  */
		switch_mach = ( 
			localstr && 
		 	 strncmp(localstr,ip1,strlen(localstr)) &&
			!strncmp(localstr,ip2,strlen(localstr)) );
		if (switch_mach) {
		
			fwrite (t->key+4, 1, 4, outfile);   /* 2nd ip  */
			fwrite (t->key  , 1, 4, outfile);   /* 1st ip  */
			fwrite (t->key+12,1, 1, outfile);   /* protocol */
			fwrite (t->key+10,1, 2, outfile);   /* 2nd port  */
			fwrite (t->key+8, 1, 2, outfile);   /* 1st port  */
			fwrite (&data->nbyte2, 8, 1, outfile);  /* 2nd ip, bytes received  */ 
			fwrite (&data->nbyte1, 8, 1, outfile);  /* 1st ip, bytes received  */
			fwrite (&data->npkt2,  4, 1, outfile);  /* 2nd ip, packets recevied  */
			fwrite (&data->npkt1,  4, 1, outfile);  /* 1st ip, packets received  */

		} else {
		
			fwrite (t->key  , 1, 4, outfile);   /* 1st ip  */
			fwrite (t->key+4, 1, 4, outfile);   /* 2nd ip  */
			fwrite (t->key+12,1, 1, outfile);   /* protocol */
			fwrite (t->key+8, 1, 2, outfile);   /* 1st port  */
			fwrite (t->key+10,1, 2, outfile);   /* 2nd port  */
			fwrite (&data->nbyte1, 8, 1, outfile); /* 1st ip, bytes received  */
			fwrite (&data->nbyte2, 8, 1, outfile); /* 2nd ip, bytes received  */
			fwrite (&data->npkt1,  4, 1, outfile); /* 1st ip, packets received  */
			fwrite (&data->npkt2,  4, 1, outfile); /* 2nd ip, packets recevied  */
		}

		/*  If switching machine order, correct first/last machine id  */
		if (write_time_m) {
			if (switch_mach) {
				first_mach  = 3 - data->time.first_mach;
				last_mach   = 3 - data->time.last_mach;
			} else { 
				first_mach  = data->time.first_mach;
				last_mach   = data->time.last_mach;
			}
			fwrite (&first_mach, sizeof(first_mach), 1, outfile);
			fwrite (&last_mach, sizeof(last_mach), 1, outfile);
		}
	}

	/*  Close file  */
	if (outname)
		fclose(outfile);
}


/*
Retrieve and print packets from hash table in text format
.. sort by time if writing it
*/
void writepkttxt (hlist_t **ha, char *outname, char *localstr) {
	hlist_t *t;
	data_t  *data;
	FILE    *outfile = stdout;
	char    ip1[16], ip2[16];
	int     pt1, pt2, prot;
	int     hour,min,sec,msec;
	int     switch_mach;
	int     first_mach, last_mach;
	int     iconn, nconn;
	hlist_t **conn = NULL;
	char    eth1str[13], eth2str[13];

	if (outname && strcmp("-",outname)) 
		outfile = fopen (outname, "wt");

	if (NULL==outfile) {
		fprintf (stderr, "ERROR:  Cannot open output file <%s>\n", outname);
		exit(1);
	}

	/*  Get number of connections  */
	nconn = getcount();
	conn = (hlist_t **) calloc (nconn, sizeof(hlist_t *));
	if (NULL==conn) {
		fprintf (stderr, 
			"ERROR:  Cannot allocate memory for connection index\n");
		exit(1);
	}

	/*  Make pointer list  */
	getfirst (ha);
	for (iconn=0;iconn<nconn;iconn++) {
		conn[iconn] = getnext(ha);
	}

	/*  Sort pointer list  */
	if (write_time_m) {
		qsort ( (void *) conn, nconn, sizeof(conn[0]), cmptime);
	} else {
		qsort ( (void *) conn, nconn, sizeof(conn[0]), cmpip  );
	}

	/*  Walk list  */
	for (iconn=0;iconn<nconn;iconn++) {
		t = conn[iconn];

		/*  Get ip addresses and ports  */
		sprintf (ip1, "%03u.%03u.%03u.%03u", 
			t->key[0], t->key[1], t->key[2], t->key[3]);
		sprintf (ip2, "%03u.%03u.%03u.%03u", 
			t->key[4], t->key[5], t->key[6], t->key[7]);
		pt1  = (int) t->key[ 8]*256 + t->key[ 9];
		pt2  = (int) t->key[10]*256 + t->key[11];
		prot = t->key[12];

		/*  Re-order ip addresses if 2nd is local and first is not  */
		switch_mach = (
			localstr && 
		 	 strncmp(localstr,ip1,strlen(localstr)) &&
			!strncmp(localstr,ip2,strlen(localstr)) );

		if (switch_mach) {
		
			/*  Print key info  */
			fprintf (outfile, "%s %s %u %u %u", ip2, ip1, prot, pt2, pt1);

			/*  Data  */
			data = (data_t *) t->data;
			fprintf (outfile, "  %d %d %d %d", 
				data->nbyte2, data->nbyte1, data->npkt2, data->npkt1);

		} else {
		
			/*  Print key info  */
			fprintf (outfile, "%s %s %u %u %u", ip1, ip2, prot, pt1, pt2);

			/*  Data  */
			data = (data_t *) t->data;
			fprintf (outfile, "  %d %d %d %d", 
				data->nbyte1, data->nbyte2, data->npkt1, data->npkt2);
		}

		if (write_time_m) {
			/*  Convert seconds from midnight to 24 hour time  */
			msec  = data->time.first_time;
			HMS(hour,min,sec,msec)
			fprintf (outfile, " %02d:%02d:%02d.%04d", hour,min,sec,msec);
			msec  = data->time.last_time;
			HMS(hour,min,sec,msec)
			fprintf (outfile, " %02d:%02d:%02d.%04d", hour,min,sec,msec);

		/*  If switching machine order, correct first/last machine id  */
			if (switch_mach) {
				first_mach  = 3 - data->time.first_mach;
				last_mach   = 3 - data->time.last_mach;
			} else { 
				first_mach  = data->time.first_mach;
				last_mach   = data->time.last_mach;
			}

			fprintf (outfile, " %1d %1d", first_mach, last_mach);
		} 

		/*  Print optional ethernet addresses  */
		if (printeth_m) {
			sprintf (eth1str, " %02x%02x%02x%02x%02x%02x", 
				t->key[13], t->key[14], t->key[15], 
				t->key[16], t->key[17], t->key[18]);
			sprintf (eth2str, " %02x%02x%02x%02x%02x%02x", 
				t->key[19], t->key[20], t->key[21], 
				t->key[22], t->key[23], t->key[24]);

			if (switch_mach) fprintf (outfile, " %s %s", eth2str, eth1str);
			else             fprintf (outfile, " %s %s", eth1str, eth2str);
		}


		fprintf (outfile, "\n");

	}

	/*  Close file  */
	if (outname)
		fclose(outfile);

	/*  reclaim storage  */
	if (NULL!=conn) free (conn);
	
}


void PrintUsage(void) {
	printf ("\nUsage: ipaudit -HLceilmoprstvwx [interface]\n");
	printf ("  Read and record info on ip connections and optionally\n");
	printf ("  dump packets to file\n");
	printf ("\n");
	printf ("  -r readfile   -  Read packets from pcap format file\n");
	printf ("                   Don't need interface with this option\n");
	printf ("  -w writefile  -  Dump selected packets to pcap format file\n");
	printf ("  -i pidfile    -  Write process id to file\n");
	printf ("  -l net-ip     -  Order output ip address pairs by network ip\n");
	printf ("  -s nlen       -  Dump first <nlen> bytes of each packet (default %d, min %d)\n",
		PLEN_DEF, PLEN_MIN);
	printf ("  -x program    -  Run program when done\n");
	printf ("  -c npacket    -  Only read in specific number of ip packets\n");
	printf ("  -e            -  Write out ethernet addresses\n");
	printf ("  -t             - Write out connection start and stop times\n");
	printf ("  -o outfile    -  Place output in 'outfile'. Default is stdout\n");
	printf ("  -p string     -  Dump only selected ip protocols and ports\n");
	printf ("     string format -p n:n:n,p,p:n where n is protocol number\n");
	printf ("     and p is port number (only for protocols 6 (tcp) and 17 (udp)\n");
	printf ("  -v            -  Print version and exit.\n");
	printf ("  -b            -  Write output in binary format (experimental)\n");
	printf ("  -t            -  Write connection start/end time\n");
	printf ("  -f filterstr  -  Use pcap filters (see tcpdump)\n");
	printf ("  -H            -  Store hosts only (protocol, ports set to zero)\n");
	printf ("  -I ipaddr     -  Dump all packets with 'ipaddr'\n");
	printf ("  -L hostportlimit,hostlimit\n");
	printf ("                -  Max number of hostport,host packets recorded\n");   
	printf ("  -N            -  Number of hash slots\n");
	printf ("Example:\n");
	printf ("  ipaudit -w dump.fil -p1:2:6,21,23  eth0\n");
	printf ("Write only packets with protocols 1 (icmp), 2 (?), and 6 (tcp)\n");
	printf ("and tcp ports 21 and 23 (ftp,telnet)\n\n");
}



/*
Determine protocols and ports to accept from string of format
	"1:6,21,23,6667:17" means protocols 1 (icmp), 6 (tcp) and 17 (udp)
	For tcp only ports 21,23,6667.
	For udp all ports (default)
*/
void parse_portstr(char *str) {
	char *p;
	char delim;
	int  lastprot = -1;
	int  is_prot;
	int  val;

	is_prot = TRUE;
	delim=' ';
	while (delim) {
		/*  Find : or '\0'  */
		p=str;
		/*  Find next : , '\0'  */
		while (*p!=':' && *p!=',' && *p!='\0') 
		p++;
		delim = *p;
		*p = '\0';
		val = atoi(str);
		str = p+1;

		if (is_prot) {
		lastprot = val;
		add_protocol(val);
		} else {
		add_port (lastprot,val);
		}
		is_prot = (delim==':');
	}
}


void add_protocol(int val) {
	/*  Check for valid protocol  */
	if (val<0 || val>255)
		return;
	/*  ALlocate if not done already  */
	if (NULL==prots_m)
		prots_m = (u_char *) alloc_err(256, sizeof(u_char));
	if (NULL==prots_m) {
		fprintf (stderr, "ERROR: Out of memory\n");
		exit (1);
	}
	/* Accept all ports  */
	prots_m[val] = PROT_ACC_ALL;
	}


void add_port (int prot, int port) {
	/*  Check for valid port  */
	if (prot<0 || prot>0x0ffff)
		return;
	/*  Check for valid protocol  */
	if (prot==PROT_TCP) {
		prots_m[prot] = PROT_ACC_SOME;
		if (NULL==tcp_ports_m) 
		tcp_ports_m = (u_char *) alloc_err(0x010000, sizeof(u_char));
		tcp_ports_m[port] = 1;
	} else if (prot==PROT_UDP) {
		prots_m[prot] = PROT_ACC_SOME;
		if (NULL==udp_ports_m) 
		udp_ports_m = (u_char *) alloc_err(0x010000, sizeof(u_char));
		tcp_ports_m[port] = 1;
	}
}



void *alloc_err (int n, int size) {
	void *p;
	p = calloc (n, size);
	if (NULL==p) {
		fprintf (stderr, "ERROR: Out of memory.\n");
		exit(1);
	}
	return p;
}


/*  
Read and format ip address :
  for example,  "112.51.1" -> "112.051.001"
*/
char *get_localstr(char *instr) {
	/*  Room for string containing formated ip address  */
	char *buffer = (char *) calloc(16,sizeof(char)); 
	char *b = buffer;
	char triplet[3];
	char *p = instr;
	int  i,n;

	/*  Search over all bytes (things between ..)  */
	while (1) {
		n = 0;
		while ((*p>='0' && *p<='9') && n<3) 
			triplet[n++] = *p++;
		/*  Either non-digit or two many digits - return NULL */
		if (*p!='.' && *p!=0) return NULL;
		for (i=n;i<3;i++) *b++ = '0';
		for (i=0;i<n;i++) *b++ = triplet[i];
		*b++ = *p;
		if (*p==0)	return buffer;
		p++;
	}
}


/*
Convert time from pcap pkthdr format (double longs) to 
1/10,000 seconds since midnight
*/
int get_pkttime (struct pcap_pkthdr *pkthdr) {
	struct tm *time;
	time = localtime( (time_t *) &pkthdr->ts.tv_sec);
	return  
		(int)
		pkthdr->ts.tv_usec/100 +  /*  Convert microseconds */
		M0SEC * (time->tm_sec + 60*(time->tm_min + 60*time->tm_hour));
}



/*  Comparison function for sorting by time  */
int cmptime (const void *ai, const void *bi) {
	hlist_t *ah = *(hlist_t **) ai;
	hlist_t *bh = *(hlist_t **) bi;
	data_t  *ad = (data_t *) (ah->data);
	data_t  *bd = (data_t *) (bh->data);
	if (ad->time.first_time < bd->time.first_time) return -1;
	if (ad->time.first_time > bd->time.first_time) return  1;
	if (ad->time.last_time  < bd->time.last_time ) return -1;
	if (ad->time.last_time  > bd->time.last_time ) return  1;
	return 0;
}


/*  Comparison function for sorting by ip packet keys  */
int cmpip  (const void *ai, const void *bi) {

	hlist_t *ah = *(hlist_t **) ai;
	hlist_t *bh = *(hlist_t **) bi;
	return memcmp (ah->key, bh->key, 13);
}


int get_packetoffset (int DataLinkType) {
	int PacketOffset;
	switch (DataLinkType) {
		case DLT_EN10MB:
		case DLT_IEEE802:
			PacketOffset = POFF_ETH;
			break;
		case DLT_PPP:
			PacketOffset = POFF_PPP;
			break;
		case DLT_RAW:
			PacketOffset = POFF_RAW;
			break;
		case DLT_NULL:
			PacketOffset = POFF_NULL;
			break;
		/*  Currently only know ethernet, ppp, for others we guess  */
		default:
			PacketOffset = 0;
	}
	return PacketOffset;
}
